import os
import yaml
from elasticsearch import Elasticsearch
import elasticsearch.helpers

from bel.Config import config

import logging
log = logging.getLogger(__name__)

cur_dir_name = os.path.dirname(os.path.realpath(__file__))
mappings_terms_fn = f'{cur_dir_name}/es_mappings_terms.yml'
terms_alias = 'terms'


def get_all_index_names(es):
    """Get all index names"""

    indices = es.indices.get_alias()

    return indices


def add_index_alias(es, index_name, alias_name):
    """Add index alias to index_name"""

    es.indices.put_alias(index=index_name, name=terms_alias)


def index_exists(es, index_name: str):
    """Does index exist?

    Args:
        index_name:  index to check for existence
    """
    return es.indices.exists(index=index_name)


def delete_index(es, index_name: str):
    """Delete the terms index"""

    if not index_name:
        log.warn('No index name given to delete')
        return None

    result = es.indices.delete(index=index_name)
    return result


def create_terms_index(es, index_name: str):
    """Create terms index"""

    with open(mappings_terms_fn, 'r') as f:
        mappings_terms = yaml.load(f)

    try:
        es.indices.create(index=index_name, body=mappings_terms)

    except Exception as e:
        log.error(f'Could not create elasticsearch terms index: {e}')


def delete_terms_indexes(es, index_name: str = 'terms_*'):
    """Delete all terms indexes"""

    try:
        es.indices.delete(index=index_name)
    except Exception as e:
        log.error(f'Could not delete all terms indices: {e}')


def get_client():
    """Get elasticsearch client

    Returns:
        es: Elasticsearch client handle
    """

    es = Elasticsearch([config['bel_api']['servers']['elasticsearch']], send_get_body_as='POST')

    return es


def bulk_load_docs(es, docs):
    """Bulk load docs

    Args:
        es: elasticsearch handle
        docs: Iterator of doc objects - includes index_name
    """

    chunk_size = 200

    try:
        results = elasticsearch.helpers.bulk(es, docs, chunk_size=chunk_size)
        log.debug(f'Elasticsearch documents loaded: {results[0]}')

        # elasticsearch.helpers.parallel_bulk(es, terms, chunk_size=chunk_size, thread_count=4)
        if len(results[1]) > 0:
            log.error('Bulk load errors {}'.format(results))
    except elasticsearch.ElasticsearchException as e:
        log.error('Indexing error: {}\n'.format(e))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Usage:  program.py <customer>

"""

from typing import Mapping, Any, List, MutableSequence
import copy
import json
import sys
import os

import bel.lang.belobj
import bel.lang.bel_specification
import bel.db.arangodb as arangodb
import bel.utils as utils


from bel.Config import config

import structlog
log = structlog.getLogger(__name__)

Edges = MutableSequence[Mapping[str, Any]]

arango_client = arangodb.get_client()
edgestore_db = arangodb.get_edgestore_handle(arango_client)


def nanopub_to_edges(nanopub: dict = {}, rules: List[str] = [], orthologize_targets: list = []):
    """Process nanopub into edges and load into EdgeStore

    Args:
        nanopub: BEL Nanopub
        rules: list of compute rules to process
        orthologize_targets: list of species in TAX:<int> format

    Returns:
        list: of edges

    Edge object:
        {
            "edge": {
                "subject": {
                    "name": subj_canon,
                    "name_lc": subj_canon.lower(),
                    "label": subj_lbl,
                    "label_lc": subj_lbl.lower(),
                    "components": subj_components,
                },
                "relation": {  # relation _key is based on a hash
                    "relation": edge_ast.bel_relation,
                    "edge_hash": edge_hash,
                    "edge_dt": edge_dt,
                    "nanopub_url": nanopub_url,
                    "nanopub_id": nanopub_id,
                    "citation": citation,
                    "subject_canon": subj_canon,
                    "subject": subj_lbl,
                    "object_canon": obj_canon,
                    "object": obj_lbl,
                    "annotations": nanopub['annotations'],
                    "metadata": nanopub['metadata'],
                    "public_flag": True,  # will be added when groups/permissions feature is finished,
                    "edge_types": edge_types,
                },
                'object': {
                    "name": obj_canon,
                    "name_lc": obj_canon.lower(),
                    "label": obj_lbl,
                    "label_lc": obj_lbl.lower(),
                    "components": obj_components,
                }
            }
        }
    """

    # Collect input values ####################################################
    nanopub_url = nanopub.get('source_url', '')

    edge_dt = utils.dt_utc_formatted()  # don't want this in relation_id

    # Extract BEL Version and make sure we can process this
    if nanopub['nanopub']['type']['name'].upper() == "BEL":
        bel_version = nanopub['nanopub']['type']['version']
        versions = bel.lang.bel_specification.get_bel_versions()
        if bel_version not in versions:
            log.error(f'Do not know this BEL Version: {bel_version}, these are the ones I can process: {versions.keys()}')
            return []
    else:
        log.error(f"Not a BEL Nanopub according to nanopub.type.name: {nanopub['nanopub']['type']['name']}")
        return []

    # Required for BEL parsing/canonicalization/orthologization
    api_url = config['bel_api']['servers']['api_url']

    try:
        citation_string = normalize_nanopub_citation(nanopub)
    except Exception as e:
        log.error(f'Could not create citation string for {nanopub_url}')
        citation_string = ''

    if orthologize_targets == []:
        if config['bel_api'].get('edges', None):
            orthologize_targets = config['bel_api']['edges'].get('orthologize_targets', [])

    # orig_species_id = [anno['id'] for anno in nanopub['nanopub']['annotations'] if anno['type'] == 'Species']
    # if orig_species_id:
    #     orig_species_id = orig_species_id[0]

    master_annotations = copy.deepcopy(nanopub['nanopub']['annotations'])
    master_metadata = copy.deepcopy(nanopub['nanopub']['metadata'])
    master_metadata.pop('gd:abstract', None)

    nanopub_type = nanopub['nanopub']['metadata'].get('nanopub_type')

    # Create Edge Assertion Info ##############################################
    # r = generate_assertion_edge_info(nanopub['nanopub']['assertions'], orig_species_id, orthologize_targets, bel_version, api_url, nanopub_type)
    r = generate_assertion_edge_info(nanopub['nanopub']['assertions'], orthologize_targets, bel_version, api_url, nanopub_type)
    edge_info_list = r['edge_info_list']

    # Build Edges #############################################################
    edges = []
    errors = []
    for edge_info in edge_info_list:
        annotations = copy.deepcopy(master_annotations)
        metadata = copy.deepcopy(master_metadata)

        errors.extend(edge_info['errors'])

        if not edge_info.get('canonical'):
            continue

        # TODO - remove this
        # if edge_info.get('species_id', False):
        #     annotations = orthologize_context(edge_info['species_id'], annotations)

        edge_hash = utils._create_hash(f'{edge_info["canonical"]["subject"]} {edge_info["canonical"]["relation"]} {edge_info["canonical"]["object"]}')

        edge = {
            'edge': {
                'subject': {
                    "name": edge_info['canonical']['subject'],
                    "name_lc": edge_info['canonical']['subject'].lower(),
                    "label": edge_info['decanonical']['subject'],
                    "label_lc": edge_info['decanonical']['subject'].lower(),
                    "components": edge_info['subject_comp'],
                },
                'relation': {
                    'relation': edge_info['canonical']['relation'],
                    'edge_hash': edge_hash,
                    'edge_dt': edge_dt,
                    'nanopub_url': nanopub_url,
                    'nanopub_id': nanopub['nanopub']['id'],
                    'citation': citation_string,
                    'subject_canon': edge_info['canonical']['subject'],
                    'subject': edge_info['decanonical']['subject'],
                    'object_canon': edge_info['canonical']['object'],
                    'object': edge_info['decanonical']['object'],
                    'annotations': copy.deepcopy(annotations),
                    'metadata': copy.deepcopy(metadata),
                    'public_flag': True,
                    'edge_types': edge_info['edge_types'],
                    'species_id': edge_info['species_id'],
                    'species_label': edge_info['species_label'],
                },
                'object': {
                    "name": edge_info['canonical']['object'],
                    "name_lc": edge_info['canonical']['object'].lower(),
                    "label": edge_info['decanonical']['object'],
                    "label_lc": edge_info['decanonical']['object'].lower(),
                    "components": edge_info['object_comp'],
                }
            }
        }

        edges.append(copy.deepcopy(edge))

    return {"edges": edges, "nanopub_id": nanopub['nanopub']['id'], "nanopub_url": nanopub_url, "success": True, "errors": errors}


def extract_ast_species(ast):
    """Extract species from ast.species set of tuples (id, label)"""

    species_id = 'None'
    species_label = 'None'
    species = [(species_id, species_label,) for (species_id, species_label) in ast.species if species_id]

    if len(species) == 1:
        (species_id, species_label) = species[0]

    if not species_id:
        species_id = 'None'
        species_label = 'None'

    log.debug(f'AST Species: {ast.species}  Species: {species}  SpeciesID: {species_id}')

    return (species_id, species_label)


# def generate_assertion_edge_info(assertions: List[dict], orig_species_id: str, orthologize_targets: List[str], bel_version: str, api_url: str, nanopub_type: str) -> dict:
def generate_assertion_edge_info(assertions: List[dict], orthologize_targets: List[str], bel_version: str, api_url: str, nanopub_type: str = '') -> dict:
    """Create edges (SRO) for assertions given orthologization targets

    Args:
        assertions: list of BEL statements (SRO object)
        orthologize_targets: list of species in TAX:<int> format
        bel_version: to be used for processing assertions
        api_url: BEL API url endpoint to use for terminologies and orthologies
    """

    bo = bel.lang.belobj.BEL(bel_version, api_url)
    bo_computed = bel.lang.belobj.BEL(bel_version, api_url)

    edge_info_list = []

    with utils.Timer() as t:
        for assertion in assertions:
            # if not assertion.get('relation', False):
            #     continue  # Skip any subject only statements

            start_time = t.elapsed
            bo.parse(assertion)

            if not bo.ast:
                errors = [f'{error[0]} {error[1]}' for error in bo.validation_messages if error[0] == 'ERROR']
                edge_info = {'errors': copy.deepcopy(errors)}
                edge_info_list.append(copy.deepcopy(edge_info))
                continue

            # populate canonical terms and orthologs for assertion
            bo.collect_nsarg_norms()
            bo.collect_orthologs(orthologize_targets)

            log.info('Timing - time to collect nsargs and orthologs', delta_ms=(t.elapsed - start_time))

            (edge_species_id, edge_species_label) = extract_ast_species(bo.ast)
            orig_species_id = edge_species_id

            canon = bo.canonicalize().to_triple()

            components = get_node_subcomponents(bo.ast)  # needs to be run after canonicalization
            computed_asts = bo.compute_edges(ast_result=True)  # needs to be run after canonicalization
            decanon = bo.decanonicalize().to_triple()

            if nanopub_type == 'backbone':
                edge_types = ['backbone']
            else:
                edge_types = ['original', 'primary']

            if assertion.get('relation', False):

                causal_edge_type = []
                if 'causal' in bo.spec['relations']['info'][assertion['relation']]['categories']:
                    edge_types.append('causal')

                edge_info = {
                    'edge_types': edge_types,
                    'species_id': edge_species_id,
                    'species_label': edge_species_label,
                    'canonical': canon,
                    'decanonical': decanon,
                    'subject_comp': components['subject_comp'],
                    'object_comp': components['object_comp'],
                    'errors': [],
                }
                edge_info_list.append(copy.deepcopy(edge_info))

            # Loop through primary computed asts
            for computed_ast in computed_asts:
                bo_computed.ast = computed_ast
                bo_computed.collect_nsarg_norms()
                canon = bo_computed.canonicalize().to_triple()
                components = get_node_subcomponents(bo_computed.ast)  # needs to be run after canonicalization
                decanon = bo_computed.decanonicalize().to_triple()

                (edge_species_id, edge_species_label) = extract_ast_species(bo_computed.ast)

                edge_info = {
                    'edge_types': ['computed'],
                    'species_id': edge_species_id,
                    'species_label': edge_species_label,
                    'canonical': canon,
                    'decanonical': decanon,
                    'subject_comp': components['subject_comp'],
                    'object_comp': components['object_comp'],
                    'errors': [],
                }
                if [edge for edge in edge_info_list if edge.get('canonical', {}) == canon]:
                    continue  # skip if edge is already included (i.e. the primary is same as computed edge)
                edge_info_list.append(copy.deepcopy(edge_info))

            # Skip orthologs if backbone nanopub
            if nanopub_type == 'backbone':
                continue

            # only process orthologs if there are species-specific NSArgs
            if len(bo.ast.species) > 0:
                # Loop through orthologs
                for species_id in orthologize_targets:
                    log.debug(f'Orig species: {orig_species_id}  Target species: {species_id}')
                    if species_id == orig_species_id:
                        continue

                    bo.orthologize(species_id)

                    (edge_species_id, edge_species_label) = extract_ast_species(bo.ast)

                    if edge_species_id == 'None' or edge_species_id == orig_species_id:
                        log.debug(f'Skipping orthologization- species == "None" or {orig_species_id}  ASTspecies: {bo.ast.species} for {bo}')
                        continue

                    ortho_decanon = bo.orthologize(species_id).to_triple()  # defaults to decanonicalized orthologized form
                    ortho_canon = bo.canonicalize().to_triple()
                    computed_asts = bo.compute_edges(ast_result=True)  # needs to be run after canonicalization
                    components = get_node_subcomponents(bo.ast)  # needs to be run after canonicalization

                    if assertion.get('relation', False):
                        edge_info = {
                            'edge_types': ['orthologized', 'primary'] + causal_edge_type,
                            'species_id': edge_species_id,
                            'species_label': edge_species_label,
                            'canonical': ortho_canon,
                            'decanonical': ortho_decanon,
                            'subject_comp': components['subject_comp'],
                            'object_comp': components['object_comp'],
                            'errors': [],
                        }

                        edge_info_list.append(copy.deepcopy(edge_info))

                    # Loop through orthologized computed asts
                    for computed_ast in computed_asts:
                        bo_computed.ast = computed_ast
                        bo_computed.collect_nsarg_norms()
                        canon = bo_computed.canonicalize().to_triple()
                        components = get_node_subcomponents(bo_computed.ast)  # needs to be run after canonicalization
                        decanon = bo_computed.decanonicalize().to_triple()

                        (edge_species_id, edge_species_label) = extract_ast_species(bo_computed.ast)

                        edge_info = {
                            'edge_types': ['computed', 'orthologized'],
                            'species_id': edge_species_id,
                            'species_label': edge_species_label,
                            'canonical': canon,
                            'decanonical': decanon,
                            'subject_comp': components['subject_comp'],
                            'object_comp': components['object_comp'],
                            'errors': [],
                        }
                        if [edge for edge in edge_info_list if edge.get('canonical', {}) == canon]:
                            continue  # skip if edge is already included (i.e. the primary is same as computed edge)
                        edge_info_list.append(copy.deepcopy(edge_info))

    log.info('Timing - Generated all edge info for all nanopub assertions', delta_ms=t.elapsed)

    return {'edge_info_list': edge_info_list}


def get_node_subcomponents(ast):

    sub, obj = [], []

    try:
        sub = ast.bel_subject.subcomponents(subcomponents=[])

        # TODO - update handling nested BEL statement - see Natalie's recommendation on how to handle
        obj_components = []
        if ast.bel_object.__class__.__name__ == 'BELAst':  # Nested BEL Assertion
            obj_components = ast.bel_object.bel_subject.subcomponents(subcomponents=[])
            obj_components = ast.bel_object.bel_object.subcomponents(subcomponents=obj_components)
        elif hasattr(ast.bel_object, 'subcomponents'):  # Normal BEL Assertion
            obj_components = ast.bel_object.subcomponents(subcomponents=[])

        obj = obj_components

    except Exception as e:
        log.warning(f'Problem getting subcomponents for {ast.to_string()}', error=str(e))

    return {'subject_comp': sub, 'object_comp': obj}


def orthologize_context(orthologize_target: str, annotations: Mapping[str, Any]) -> Mapping[str, Any]:
    """Orthologize context

    Replace Species context with new orthologize target and add a annotation type of OrthologizedFrom
    """

    url = f'{config["bel_api"]["servers"]["api_url"]}/terms/{orthologize_target}'
    r = utils.get_url(url)
    species_label = r.json().get("label", "unlabeled")

    orthologized_from = {}
    for idx, annotation in enumerate(annotations):
        if annotation['type'] == 'Species':
            orthologized_from = {'id': annotation['id'], 'label': annotation['label']}
            annotations[idx] = {'type': 'Species', 'id': orthologize_target, 'label': species_label}

    if 'id' in orthologized_from:
        annotations.append({'type': 'OrigSpecies', 'id': f'Orig-{orthologized_from["id"]}', 'label': f'Orig-{orthologized_from["label"]}'})

    return annotations


def normalize_nanopub_citation(nanopub):

    citation_string = ''
    if 'database' in nanopub['nanopub']['citation']:
        if nanopub['nanopub']['citation']['database']['name'].lower() == 'pubmed':
            citation_string = f"PMID:{nanopub['nanopub']['citation']['database']['id']}"
        else:
            citation_string = f"{nanopub['nanopub']['citation']['database']['name']}:{nanopub['nanopub']['citation']['database']['id']}"
    elif 'reference' in nanopub['nanopub']['citation'] and isinstance(nanopub['nanopub']['citation']['reference'], str):
        citation_string = nanopub['nanopub']['citation']['reference']
    elif 'uri' in nanopub['nanopub']['citation']:
        citation_string = nanopub['nanopub']['citation']['uri']

    return citation_string


def main():
    pass


if __name__ == '__main__':
    # Setup logging
    import logging.config
    module_fn = os.path.basename(__file__)
    module_fn = module_fn.replace('.py', '')

    if config.get('logging', False):
        logging.config.dictConfig(config.get('logging'))

    log = logging.getLogger(f'{module_fn}')

    main()



import timy
import json
import gzip
import copy

from arango import ArangoError

from typing import IO

import bel.utils
import bel.db.elasticsearch as elasticsearch
import bel.db.arangodb as arangodb

from bel.Config import config

# import structlog
# import logging
# log = logging.getLogger(__name__)

from structlog import get_logger
log = get_logger()

terms_alias = 'terms'


def load_terms(fo: IO, metadata: dict, forceupdate: bool):
    """Load terms into Elasticsearch and ArangoDB

    Forceupdate will create a new index in Elasticsearch regardless of whether
    an index with the resource version already exists.

    Args:
        fo: file obj - terminology file
        metadata: dict containing the metadata for terminology
        forceupdate: force full update - e.g. don't leave Elasticsearch indexes
            alone if their version ID matches
    """

    version = metadata['metadata']['version']

    # LOAD TERMS INTO Elasticsearch
    with timy.Timer('Load Terms') as timer:
        es = bel.db.elasticsearch.get_client()

        es_version = version.replace('T', '').replace('-', '').replace(':', '')
        index_prefix = f"terms_{metadata['metadata']['namespace'].lower()}"
        index_name = f"{index_prefix}_{es_version}"

        # Create index with mapping
        if not elasticsearch.index_exists(es, index_name):
            elasticsearch.create_terms_index(es, index_name)
        elif forceupdate:  # force an update to the index
            index_name += '_alt'
            elasticsearch.create_terms_index(es, index_name)
        else:
            return  # Skip loading if not forced and not a new namespace

        terms_iterator = terms_iterator_for_elasticsearch(fo, index_name)
        elasticsearch.bulk_load_docs(es, terms_iterator)

        # Remove old namespace index
        index_names = elasticsearch.get_all_index_names(es)
        for name in index_names:
            if name != index_name and index_prefix in name:
                elasticsearch.delete_index(es, name)

        # Add terms_alias to this index
        elasticsearch.add_index_alias(es, index_name, terms_alias)

        log.info('Load namespace terms', elapsed=timer.elapsed, namespace=metadata['metadata']['namespace'])

    # LOAD EQUIVALENCES INTO ArangoDB
    with timy.Timer('Load Term Equivalences') as timer:
        arango_client = arangodb.get_client()
        belns_db = arangodb.get_belns_handle(arango_client)
        arangodb.batch_load_docs(belns_db, terms_iterator_for_arangodb(fo, version), on_duplicate='update')

        log.info('Loaded namespace equivalences', elapsed=timer.elapsed, namespace=metadata['metadata']['namespace'])

        # Clean up old entries
        remove_old_equivalence_edges = f'''
            FOR edge in equivalence_edges
                FILTER edge.source == "{metadata["metadata"]["namespace"]}"
                FILTER edge.version != "{version}"
                REMOVE edge IN equivalence_edges
        '''
        remove_old_equivalence_nodes = f'''
            FOR node in equivalence_nodes
                FILTER node.source == "{metadata["metadata"]["namespace"]}"
                FILTER node.version != "{version}"
                REMOVE node IN equivalence_nodes
        '''
        arangodb.aql_query(belns_db, remove_old_equivalence_edges)
        arangodb.aql_query(belns_db, remove_old_equivalence_nodes)

    # Add metadata to resource metadata collection
    metadata['_key'] = f"Namespace_{metadata['metadata']['namespace']}"
    try:
        belns_db.collection(arangodb.belns_metadata_name).insert(metadata)
    except ArangoError as ae:
        belns_db.collection(arangodb.belns_metadata_name).replace(metadata)


def terms_iterator_for_arangodb(fo, version):

    species_list = config['bel_resources'].get('species_list', [])

    fo.seek(0)
    with gzip.open(fo, 'rt') as f:
        for line in f:
            term = json.loads(line)
            # skip if not term record (e.g. is a metadata record)
            if 'term' not in term:
                continue
            term = term['term']

            species_id = term.get('species_id', None)
            # Skip if species not listed in species_list
            if species_list and species_id and species_id not in species_list:
                continue

            source = term['namespace']
            term_id = term['id']
            term_key = arangodb.arango_id_to_key(term_id)

            (ns, val) = term_id.split(':', maxsplit=1)

            # Add primary ID node
            yield (arangodb.equiv_nodes_name, {'_key': term_key, 'name': term_id, 'primary': True, 'namespace': ns, 'source': source, 'version': version})

            # Create Alt ID nodes/equivalences (to support other database equivalences using non-preferred Namespace IDs)
            if 'alt_ids' in term:
                for alt_id in term['alt_ids']:
                    # log.info(f'Added {alt_id} equivalence')
                    alt_id_key = arangodb.arango_id_to_key(alt_id)
                    yield (arangodb.equiv_nodes_name, {'_key': alt_id_key, 'name': alt_id, 'namespace': ns, 'source': source, 'version': version})

                    arango_edge = {
                        '_from': f"{arangodb.equiv_nodes_name}/{term_key}",
                        '_to': f"{arangodb.equiv_nodes_name}/{alt_id_key}",
                        '_key': bel.utils._create_hash(f'{term_id}>>{alt_id}'),
                        'type': 'equivalent_to',
                        'source': source,
                        'version': version,
                    }
                    yield (arangodb.equiv_edges_name, arango_edge)

            # Cross-DB equivalences
            if 'equivalences' in term:
                for eqv in term['equivalences']:
                    (ns, val) = eqv.split(':', maxsplit=1)
                    eqv_key = arangodb.arango_id_to_key(eqv)

                    yield (arangodb.equiv_nodes_name, {'_key': eqv_key, 'name': eqv, 'namespace': ns, 'source': source, 'version': version})

                    arango_edge = {
                        '_from': f"{arangodb.equiv_nodes_name}/{term_key}",
                        '_to': f"{arangodb.equiv_nodes_name}/{eqv_key}",
                        '_key': bel.utils._create_hash(f'{term_id}>>{eqv}'),
                        'type': 'equivalent_to',
                        'source': source,
                        'version': version,
                    }
                    yield (arangodb.equiv_edges_name, arango_edge)


def terms_iterator_for_elasticsearch(fo: IO, index_name: str):
    """Add index_name to term documents for bulk load"""

    species_list = config['bel_resources'].get('species_list', [])

    fo.seek(0)  # Seek back to beginning of file
    with gzip.open(fo, 'rt') as f:
        for line in f:
            term = json.loads(line)
            # skip if not term record (e.g. is a metadata record)
            if 'term' not in term:
                continue
            term = term['term']

            # Filter species if enabled in config
            species_id = term.get('species_id', None)
            if species_list and species_id and species_id not in species_list:
                continue

            all_term_ids = set()
            for term_id in [term['id']] + term.get('alt_ids', []):
                all_term_ids.add(term_id)
                all_term_ids.add(lowercase_term_id(term_id))

            term['alt_ids'] = copy.copy(list(all_term_ids))

            yield {
                '_op_type': 'index',
                '_index': index_name,
                '_type': 'term',
                '_id': term['id'],
                '_source': copy.deepcopy(term)
            }


def lowercase_term_id(term_id: str) -> str:
    """Lowercase the term value (not the namespace prefix)

    Args:
        term_id (str): term identifier with namespace prefix, e.g. MESH:Atherosclerosis

    Returns:
        str: lowercased, e.g. MESH:atherosclerosis
    """
    (ns, val) = term_id.split(':', maxsplit=1)
    term_id = f'{ns}:{val.lower()}'

    return term_id



import timy
import json
import gzip

from arango import ArangoError

from typing import IO

import bel.utils
import bel.db.arangodb as arangodb

from bel.Config import config

from structlog import get_logger
log = get_logger()


def load_orthologs(fo: IO, metadata: dict):
    """Load orthologs into ArangoDB

    Args:
        fo: file obj - orthologs file
        metadata: dict containing the metadata for orthologs
    """

    version = metadata['metadata']['version']

    # LOAD ORTHOLOGS INTO ArangoDB
    with timy.Timer('Load Orthologs') as timer:
        arango_client = arangodb.get_client()
        belns_db = arangodb.get_belns_handle(arango_client)
        arangodb.batch_load_docs(belns_db, orthologs_iterator(fo, version), on_duplicate='update')

        log.info('Load orthologs', elapsed=timer.elapsed, source=metadata['metadata']['source'])

        # Clean up old entries
        remove_old_ortholog_edges = f'''
            FOR edge in ortholog_edges
                FILTER edge.source == "{metadata["metadata"]["source"]}"
                FILTER edge.version != "{version}"
                REMOVE edge IN ortholog_edges
        '''
        remove_old_ortholog_nodes = f'''
            FOR node in ortholog_nodes
                FILTER node.source == "{metadata["metadata"]["source"]}"
                FILTER node.version != "{version}"
                REMOVE node IN ortholog_nodes
        '''
        arangodb.aql_query(belns_db, remove_old_ortholog_edges)
        arangodb.aql_query(belns_db, remove_old_ortholog_nodes)

    # Add metadata to resource metadata collection
    metadata['_key'] = f"Orthologs_{metadata['metadata']['source']}"
    try:
        belns_db.collection(arangodb.belns_metadata_name).insert(metadata)
    except ArangoError as ae:
        belns_db.collection(arangodb.belns_metadata_name).replace(metadata)


def orthologs_iterator(fo, version):
    """Ortholog node and edge iterator"""

    species_list = config['bel_resources'].get('species_list', [])

    fo.seek(0)
    with gzip.open(fo, 'rt') as f:
        for line in f:
            edge = json.loads(line)
            if 'metadata' in edge:
                source = edge['metadata']['source']
                continue

            if 'ortholog' in edge:
                edge = edge['ortholog']
                subj_tax_id = edge['subject']['tax_id']
                obj_tax_id = edge['object']['tax_id']

                # Skip if species not listed in species_list
                if species_list and subj_tax_id and subj_tax_id not in species_list:
                    continue
                if species_list and obj_tax_id and obj_tax_id not in species_list:
                    continue

                # Converted to ArangoDB legal chars for _key
                subj_key = arangodb.arango_id_to_key(edge['subject']['id'])
                subj_id = edge['subject']['id']

                # Converted to ArangoDB legal chars for _key
                obj_key = arangodb.arango_id_to_key(edge['object']['id'])
                obj_id = edge['object']['id']

                # Subject node
                yield (arangodb.ortholog_nodes_name, {'_key': subj_key, 'name': subj_id, 'tax_id': edge['subject']['tax_id'], 'source': source, 'version': version})
                # Object node
                yield (arangodb.ortholog_nodes_name, {'_key': obj_key, 'name': obj_id, 'tax_id': edge['object']['tax_id'], 'source': source, 'version': version})

                arango_edge = {
                    '_from': f"{arangodb.ortholog_nodes_name}/{subj_key}",
                    '_to': f"{arangodb.ortholog_nodes_name}/{obj_key}",
                    '_key': bel.utils._create_hash(f'{subj_id}>>{obj_id}'),
                    'type': 'ortholog_to',
                    'source': source,
                    'version': version,
                }

                yield (arangodb.ortholog_edges_name, arango_edge)

